@@ -1,23 +1,23 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*- */
 /*
  * gtksourcegutter.c
  * This file is part of gtksourceview
  *
  * Copyright (C) 2009 - Jesse van den Kieboom
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * gtksourceview is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * gtksourceview is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor,
- * Boston, MA  02110-1301  USA
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include "gtksourcegutter-private.h"
@@ -25,6 +25,37 @@
 #include "gtksourceview-i18n.h"
 #include "gtksourceview-marshal.h"
 
+/**
+ * SECTION:gutter
+ * @Short_description: Gutter object for #GtkSourceView
+ * @Title: GtkSourceGutter
+ * @See_also:#GtkSourceView
+ *
+ * The #GtkSourceGutter object represents the left and right gutters of the text
+ * view. It is used by #GtkSourceView to draw the line numbers and category
+ * marks that might be present on a line. By packing additional #GtkCellRenderer
+ * objects in the gutter, you can extend the gutter with your own custom 
+ * drawings.
+ *
+ * The gutter works very much the same way as cells rendered in a #GtkTreeView.
+ * The concept is similar, with the exception that the gutter does not have an
+ * underlying #GtkTreeModel. Instead, you should use
+ * #gtk_source_gutter_set_cell_data_func to set a callback to fill in any of the
+ * cell renderers properties, given the line for which the cell is to be
+ * rendered. Renderers are inserted into the gutter at a certain position. The
+ * builtin line number renderer is at position 
+
+ * #GTK_SOURCE_VIEW_GUTTER_POSITION_LINES (-30) and the marks renderer is at
+ * #GTK_SOURCE_VIEW_GUTTER_POSITION_MARKS (-20). You can use these values to
+ * position custom renderers accordingly. The width of a cell renderer can be
+ * specified as either fixed (using
+
+ * #gtk_cell_renderer_set_fixed_size) or dynamic, in which case you
+ * <emphasis>must</emphasis> set #gtk_source_gutter_set_cell_size_func. This
+ * callback is used to set the properties of the renderer such that
+ * #gtk_cell_renderer_get_size yields the maximum width of the cell.
+ */
+
 #define GTK_SOURCE_GUTTER_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), GTK_TYPE_SOURCE_GUTTER, GtkSourceGutterPrivate))
 
 /* Properties */
@@ -74,7 +105,7 @@ struct _GtkSourceGutterPrivate
 {
 	GtkSourceView *view;
 	GtkTextWindowType window_type;
-
+	gint size;
 	GList *renderers;
 	guint signals[LAST_EXTERNAL_SIGNAL];
 };
@@ -206,6 +237,8 @@ set_view (GtkSourceGutter *gutter,
 {
 	gutter->priv->view = view;
 
+	gutter->priv->size = -1;
+
 	g_object_weak_ref (G_OBJECT (view),
 	                   (GWeakNotify)view_notify,
 	                   gutter);
@@ -275,7 +308,7 @@ revalidate_size (GtkSourceGutter *gutter)
 		gtk_text_view_set_border_window_size (GTK_TEXT_VIEW (gutter->priv->view),
 		                                      gutter->priv->window_type,
 		                                      1);
-		do_redraw (gutter);
+		gutter->priv->size = -1;
 	}
 	else if (window && !gutter->priv->renderers)
 	{
@@ -376,7 +409,7 @@ gtk_source_gutter_class_init (GtkSourceGutterClass *klass)
 			      3,
 			      GTK_TYPE_CELL_RENDERER,
 			      GTK_TYPE_TEXT_ITER,
-			      G_TYPE_POINTER);
+			      GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
 
 	/**
 	 * GtkSourceGutter::query-tooltip:
@@ -409,6 +442,8 @@ static void
 gtk_source_gutter_init (GtkSourceGutter *self)
 {
 	self->priv = GTK_SOURCE_GUTTER_GET_PRIVATE (self);
+
+	self->priv->size = -1;
 }
 
 static gint
@@ -682,7 +717,6 @@ gtk_source_gutter_queue_draw (GtkSourceGutter *gutter)
 	revalidate_size (gutter);
 }
 
-/* Callbacks */
 static gint
 calculate_size (GtkSourceGutter  *gutter,
                 Renderer         *renderer)
@@ -831,19 +865,8 @@ on_view_expose_event (GtkSourceView   *view,
 {
 	GdkWindow *window;
 	GtkTextView *text_view;
-	gint size;
-	gint y1, y2;
-	GArray *numbers;
-	GArray *pixels;
-	GArray *heights;
 	GArray *sizes;
-	GtkTextIter cur;
-	gint cur_line;
-	gint count;
-	gint i;
-	GList *item;
-	gint x;
-	gint y;
+	gint size;
 
 	window = gtk_source_gutter_get_window (gutter);
 
@@ -852,119 +875,153 @@ on_view_expose_event (GtkSourceView   *view,
 		return FALSE;
 	}
 
-	gdk_window_get_pointer (window, &x, &y, NULL);
-
 	text_view = GTK_TEXT_VIEW (view);
 	sizes = g_array_new (FALSE, FALSE, sizeof (gint));
 
+	/* This is fairly ugly, but we could not find a better way to
+	 * do it: renderers could have changed size and they do not have
+	 * a way to signal it. So on expose we revalidate the size and
+	 * requeue another expose.
+	 * To see if the size has changed we test the size of only the
+	 * gutter itself since the full border window size is not under
+	 * our control (see e.g bug #589382).
+	 * This also means that if the user manually forces the
+	 * border_window_size to a value smaller than the gutter, things
+	 * will not be drawn properly. */
 	size = calculate_sizes (gutter, sizes);
-
-	if (gtk_text_view_get_border_window_size (text_view, gutter->priv->window_type) != size)
+	if (gutter->priv->size != size)
 	{
+
+		gint border_size;
+
+		border_size = gtk_text_view_get_border_window_size (text_view, gutter->priv->window_type);
+
+		if (gutter->priv->size < 0)
+			border_size += size;
+		else
+			border_size = MAX (0, border_size - gutter->priv->size) + size;
+
+		gutter->priv->size = size;
+
 		/* Will trigger a new expose */
-		gtk_text_view_set_border_window_size (text_view, gutter->priv->window_type, size);
-		return FALSE;
+		gtk_text_view_set_border_window_size (text_view, gutter->priv->window_type, border_size);
 	}
-
-	y1 = event->area.y;
-	y2 = y1 + event->area.height;
-
-	/* get the extents of the line printing */
-	gtk_text_view_window_to_buffer_coords (text_view,
-					       gutter->priv->window_type,
-					       0,
-					       y1,
-					       NULL,
-					       &y1);
-
-	gtk_text_view_window_to_buffer_coords (text_view,
-					       gutter->priv->window_type,
-					       0,
-					       y2,
-					       NULL,
-					       &y2);
-
-	numbers = g_array_new (FALSE, FALSE, sizeof (gint));
-	pixels = g_array_new (FALSE, FALSE, sizeof (gint));
-	heights = g_array_new (FALSE, FALSE, sizeof (gint));
-
-	/* get the line numbers and y coordinates. */
-	get_lines (text_view, y1, y2, pixels, heights, numbers, &count);
-
-	gtk_text_buffer_get_iter_at_mark (text_view->buffer,
-					  &cur,
-					  gtk_text_buffer_get_insert (text_view->buffer));
-
-	cur_line = gtk_text_iter_get_line (&cur);
-
-	for (i = 0; i < count; ++i)
+	else
 	{
-		gint pos;
-		gint line_to_paint;
-		GdkRectangle cell_area;
-		gint idx = 0;
+		gint x, y;
+		gint y1, y2;
+		GArray *numbers;
+		GArray *pixels;
+		GArray *heights;
+		GtkTextIter cur;
+		gint cur_line;
+		gint count;
+		gint i;
+		GList *item;
 
-		gtk_text_view_buffer_to_window_coords (text_view,
+		gdk_window_get_pointer (window, &x, &y, NULL);
+
+		y1 = event->area.y;
+		y2 = y1 + event->area.height;
+
+		/* get the extents of the line printing */
+		gtk_text_view_window_to_buffer_coords (text_view,
 						       gutter->priv->window_type,
 						       0,
-						       g_array_index (pixels, gint, i),
+						       y1,
 						       NULL,
-						       &pos);
+						       &y1);
 
-		line_to_paint = g_array_index (numbers, gint, i);
+		gtk_text_view_window_to_buffer_coords (text_view,
+						       gutter->priv->window_type,
+						       0,
+						       y2,
+						       NULL,
+						       &y2);
 
-		cell_area.x = 0;
-		cell_area.y = pos;
-		cell_area.height = g_array_index (heights, gint, i);
+		numbers = g_array_new (FALSE, FALSE, sizeof (gint));
+		pixels = g_array_new (FALSE, FALSE, sizeof (gint));
+		heights = g_array_new (FALSE, FALSE, sizeof (gint));
 
-		for (item = gutter->priv->renderers; item; item = g_list_next (item))
+		/* get the line numbers and y coordinates. */
+		get_lines (text_view, y1, y2, pixels, heights, numbers, &count);
+
+		gtk_text_buffer_get_iter_at_mark (text_view->buffer,
+						  &cur,
+						  gtk_text_buffer_get_insert (text_view->buffer));
+
+		cur_line = gtk_text_iter_get_line (&cur);
+
+		for (i = 0; i < count; ++i)
 		{
-			Renderer *renderer = (Renderer *)item->data;
-			gint width = g_array_index (sizes, gint, idx++);
-			GtkCellRendererState state = 0;
+			gint pos;
+			gint line_to_paint;
+			GdkRectangle cell_area;
+			gint idx = 0;
 
-			cell_area.width = width;
+			gtk_text_view_buffer_to_window_coords (text_view,
+							       gutter->priv->window_type,
+							       0,
+							       g_array_index (pixels, gint, i),
+							       NULL,
+							       &pos);
 
-			/* Call data func if needed */
-			if (renderer->data_func)
+			line_to_paint = g_array_index (numbers, gint, i);
+
+			cell_area.x = 0;
+			cell_area.y = pos;
+			cell_area.height = g_array_index (heights, gint, i);
+
+			for (item = gutter->priv->renderers; item; item = g_list_next (item))
 			{
-				renderer->data_func (gutter,
-				                     renderer->renderer,
-				                     line_to_paint,
-				                     line_to_paint == cur_line,
-				                     renderer->data_func_data);
+				Renderer *renderer = (Renderer *)item->data;
+				gint width = g_array_index (sizes, gint, idx++);
+				GtkCellRendererState state = 0;
+
+				cell_area.width = width;
+
+				/* Call data func if needed */
+				if (renderer->data_func)
+				{
+					renderer->data_func (gutter,
+						             renderer->renderer,
+						             line_to_paint,
+						             line_to_paint == cur_line,
+						             renderer->data_func_data);
+				}
+
+				if (x >= cell_area.x && x < cell_area.x + cell_area.width &&
+				    y >= cell_area.y && y < cell_area.y + cell_area.height)
+				{
+					GtkCellRendererMode mode;
+
+					g_object_get (G_OBJECT (renderer->renderer),
+						      "mode", &mode,
+						      NULL);
+
+					if (mode & GTK_CELL_RENDERER_MODE_ACTIVATABLE)
+						state = GTK_CELL_RENDERER_PRELIT;
+				}
+
+				/* Call render with correct area */
+				gtk_cell_renderer_render (renderer->renderer,
+					                  window,
+					                  GTK_WIDGET (view),
+					                  &cell_area,
+					                  &cell_area,
+					                  &cell_area,
+					                  state);
+
+				cell_area.x += cell_area.width;
 			}
-
-			if (x >= cell_area.x && x < cell_area.x + cell_area.width &&
-			    y >= cell_area.y && y < cell_area.y + cell_area.height)
-			{
-				GtkCellRendererMode mode;
-
-				g_object_get (G_OBJECT (renderer->renderer),
-				              "mode", &mode,
-				              NULL);
-
-				if (mode & GTK_CELL_RENDERER_MODE_ACTIVATABLE)
-					state = GTK_CELL_RENDERER_PRELIT;
-			}
-
-			/* Call render with correct area */
-			gtk_cell_renderer_render (renderer->renderer,
-			                          window,
-			                          GTK_WIDGET (view),
-			                          &cell_area,
-			                          &cell_area,
-			                          &cell_area,
-			                          state);
-
-			cell_area.x += cell_area.width;
 		}
+
+		g_array_free (numbers, TRUE);
+		g_array_free (pixels, TRUE);
+		g_array_free (heights, TRUE);
 	}
 
 	g_array_free (sizes, TRUE);
-	g_array_free (numbers, TRUE);
-	g_array_free (pixels, TRUE);
-	g_array_free (heights, TRUE);
 
 	return FALSE;
 }
